/***********************************************************************
 * Copyright (c) 2004,2005,2006,2007 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Actuate Corporation - initial API and implementation
 ***********************************************************************/

package org.eclipse.birt.chart.render;

import java.awt.Color;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import org.eclipse.birt.chart.computation.BoundingBox;
import org.eclipse.birt.chart.computation.DataPointHints;
import org.eclipse.birt.chart.computation.IConstants;
import org.eclipse.birt.chart.computation.Methods;
import org.eclipse.birt.chart.device.IDeviceRenderer;
import org.eclipse.birt.chart.device.IDisplayServer;
import org.eclipse.birt.chart.device.IStructureDefinitionListener;
import org.eclipse.birt.chart.engine.extension.i18n.Messages;
import org.eclipse.birt.chart.event.ArcRenderEvent;
import org.eclipse.birt.chart.event.AreaRenderEvent;
import org.eclipse.birt.chart.event.EventObjectCache;
import org.eclipse.birt.chart.event.InteractionEvent;
import org.eclipse.birt.chart.event.LineRenderEvent;
import org.eclipse.birt.chart.event.PolygonRenderEvent;
import org.eclipse.birt.chart.event.PrimitiveRenderEvent;
import org.eclipse.birt.chart.event.StructureSource;
import org.eclipse.birt.chart.event.TextRenderEvent;
import org.eclipse.birt.chart.event.WrappedStructureSource;
import org.eclipse.birt.chart.exception.ChartException;
import org.eclipse.birt.chart.log.ILogger;
import org.eclipse.birt.chart.log.Logger;
import org.eclipse.birt.chart.model.ChartWithoutAxes;
import org.eclipse.birt.chart.model.attribute.Bounds;
import org.eclipse.birt.chart.model.attribute.ChartDimension;
import org.eclipse.birt.chart.model.attribute.ColorDefinition;
import org.eclipse.birt.chart.model.attribute.Fill;
import org.eclipse.birt.chart.model.attribute.Gradient;
import org.eclipse.birt.chart.model.attribute.Insets;
import org.eclipse.birt.chart.model.attribute.LeaderLineStyle;
import org.eclipse.birt.chart.model.attribute.LegendItemType;
import org.eclipse.birt.chart.model.attribute.LineAttributes;
import org.eclipse.birt.chart.model.attribute.LineStyle;
import org.eclipse.birt.chart.model.attribute.Location;
import org.eclipse.birt.chart.model.attribute.Palette;
import org.eclipse.birt.chart.model.attribute.Position;
import org.eclipse.birt.chart.model.attribute.Size;
import org.eclipse.birt.chart.model.attribute.impl.BoundsImpl;
import org.eclipse.birt.chart.model.attribute.impl.ColorDefinitionImpl;
import org.eclipse.birt.chart.model.attribute.impl.GradientImpl;
import org.eclipse.birt.chart.model.attribute.impl.InsetsImpl;
import org.eclipse.birt.chart.model.attribute.impl.LineAttributesImpl;
import org.eclipse.birt.chart.model.attribute.impl.LocationImpl;
import org.eclipse.birt.chart.model.attribute.impl.SizeImpl;
import org.eclipse.birt.chart.model.component.Label;
import org.eclipse.birt.chart.model.component.impl.LabelImpl;
import org.eclipse.birt.chart.model.data.Trigger;
import org.eclipse.birt.chart.model.data.impl.TriggerImpl;
import org.eclipse.birt.chart.model.type.PieSeries;
import org.eclipse.birt.chart.plugin.ChartEngineExtensionPlugin;
import org.eclipse.birt.chart.script.ScriptHandler;
import org.eclipse.birt.chart.util.ChartUtil;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EcoreUtil;

/**
 * PieRenderer
 */
public final class PieRenderer
{

	static final int UNKNOWN = 0;

	private static final int LOWER = 1;

	private static final int UPPER = 2;

	private static final double LEADER_TICK_MIN_SIZE = 10; // POINTS

	private static final int LESS = -1;

	private static final int MORE = 1;

	private static final int EQUAL = 0;

	private final static double LENGTH_HINTS = 100d;

	private final Position lpDataPoint;

	private final Position lpSeriesTitle;

	private transient double dLeaderLength;

	private final LeaderLineStyle lls;

	/**
	 * Series thickness
	 */
	private transient double dThickness;

	private transient double dExplosion = 0;

	private transient String sExplosionExpression = null;

	private final Pie pie;

	private final PieSeries ps;

	private final ArrayList pieSliceList = new ArrayList( );

	/**
	 * Holds a list of deferred planes (flat and curved) to be sorted before
	 * rendering
	 */
	private final ArrayList alPlanes = new ArrayList( );

	private final Palette pa;

	private Label laDataPoint;

	private final Label laSeriesTitle;

	private final LineAttributes liaLL;

	private final LineAttributes liaEdges;

	private transient IDisplayServer xs = null;

	private transient IDeviceRenderer idr = null;

	private transient Bounds boTitleContainer = null;

	private transient Insets insCA = null;

	private transient Bounds boSetDuringComputation = null;

	private final boolean bPaletteByCategory;

	private transient boolean bBoundsAdjustedForInsets = false;

	private transient boolean bMinSliceDefined = false;

	private transient double dMinSlice = 0;

	private transient double dAbsoluteMinSlice = 0;

	private transient boolean bPercentageMinSlice = false;

	private transient boolean bMinSliceApplied = false;

	private transient int orginalSliceCount = 0;

	private transient double ratio = 0;

	private static ILogger logger = Logger.getLogger( "org.eclipse.birt.chart.engine.extension/render" ); //$NON-NLS-1$

	/**
	 * 
	 * @param cwoa
	 * @param pie
	 * @param dpha
	 * @param da
	 * @param pa
	 */
	PieRenderer( ChartWithoutAxes cwoa, Pie pie, DataPointHints[] dpha,
			double[] da, Palette pa ) throws ChartException
	{
		this.pa = pa;
		this.pie = pie;

		ps = (PieSeries) pie.getSeries( );
		sExplosionExpression = ps.getExplosionExpression( );
		dExplosion = ( ps.isSetExplosion( ) ? ps.getExplosion( ) : 0 )
				* pie.getDeviceScale( );
		dThickness = ( ( cwoa.getDimension( ) == ChartDimension.TWO_DIMENSIONAL_LITERAL ) ? 0
				: cwoa.getSeriesThickness( ) )
				* pie.getDeviceScale( );
		ratio = ps.isSetRatio( ) ? ps.getRatio( ) : 1;
		liaLL = ps.getLeaderLineAttributes( );
		if ( ps.getLeaderLineAttributes( ).isVisible( ) )
		{
			dLeaderLength = ps.getLeaderLineLength( ) * pie.getDeviceScale( );
		}
		else
		{
			dLeaderLength = 0;
		}

		liaEdges = LineAttributesImpl.create( ColorDefinitionImpl.BLACK( ),
				LineStyle.SOLID_LITERAL,
				1 );
		bPaletteByCategory = ( cwoa.getLegend( ).getItemType( ) == LegendItemType.CATEGORIES_LITERAL );

		lpDataPoint = ps.getLabelPosition( );
		lpSeriesTitle = ps.getTitlePosition( );
		laSeriesTitle = LabelImpl.copyInstance( ps.getTitle( ) );
		laSeriesTitle.getCaption( )
				.setValue( pie.getRunTimeContext( )
						.externalizedMessage( String.valueOf( ps.getSeriesIdentifier( ) ) ) ); // TBD:
		laSeriesTitle.getCaption( )
				.getFont( )
				.setAlignment( pie.switchTextAlignment( laSeriesTitle.getCaption( )
						.getFont( )
						.getAlignment( ) ) );
		// APPLY
		// FORMAT
		// SPECIFIER
		laDataPoint = LabelImpl.copyInstance( ps.getLabel( ) );
		lls = ps.getLeaderLineStyle( );

		bMinSliceDefined = cwoa.isSetMinSlice( );
		dMinSlice = cwoa.getMinSlice( );
		bPercentageMinSlice = cwoa.isMinSlicePercent( );

		double dTotal = 0;
		orginalSliceCount = da.length;
		for ( int i = 0; i < da.length; i++ )
		{
			if ( da[i] < 0 )
			{
				dTotal -= da[i]; // use negative values as absolute
			}
			else if ( !Double.isNaN( da[i] ) )
			{
				dTotal += da[i];
			}
		}

		if ( bMinSliceDefined )
		{
			if ( bPercentageMinSlice )
			{
				dAbsoluteMinSlice = dMinSlice * dTotal / 100d;
			}
			else
			{
				dAbsoluteMinSlice = dMinSlice;
			}

			double residualPos = 0;
			double residualNeg = 0;
			DataPointHints dphPos = null;
			DataPointHints dphNeg = null;
			for ( int i = 0; i < da.length; i++ )
			{
				// filter null values.
				if ( Double.isNaN( da[i] ) )
				{
					continue;
				}

				if ( Math.abs( da[i] ) >= Math.abs( dAbsoluteMinSlice ) )
				{
					pieSliceList.add( new PieSlice( da[i], dpha[i], i ) );
				}
				else
				{
					if ( da[i] >= 0 )
					{
						residualPos += da[i];
						if ( dphPos == null )
						{
							dphPos = dpha[i].getCopy( );
						}
						else
						{
							dphPos.accumulate( dpha[i].getBaseValue( ),
									dpha[i].getOrthogonalValue( ),
									dpha[i].getSeriesValue( ),
									dpha[i].getPercentileOrthogonalValue( ) );
						}
					}
					else
					{
						residualNeg += da[i];
						if ( dphNeg == null )
						{
							dphNeg = dpha[i].getCopy( );
						}
						else
						{
							dphNeg.accumulate( dpha[i].getBaseValue( ),
									dpha[i].getOrthogonalValue( ),
									dpha[i].getSeriesValue( ),
									dpha[i].getPercentileOrthogonalValue( ) );
						}

					}

				}

			}

			if ( dphPos != null )
			{
				dphPos.setBaseValue( cwoa.getMinSliceLabel( ) );
				dphPos.setIndex( orginalSliceCount );
				// neg and pos "other" slice share the same palette color
				pieSliceList.add( new PieSlice( residualPos,
						dphPos,
						orginalSliceCount ) );
				bMinSliceApplied = true;
			}
			if ( dphNeg != null )
			{
				dphNeg.setBaseValue( cwoa.getMinSliceLabel( ) );
				dphNeg.setIndex( orginalSliceCount );
				pieSliceList.add( new PieSlice( residualNeg,
						dphNeg,
						orginalSliceCount ) );
				bMinSliceApplied = true;
			}
		}
		else
		{
			for ( int i = 0; i < da.length; i++ )
			{
				// filter null values.
				if ( Double.isNaN( da[i] ) )
				{
					continue;
				}
				pieSliceList.add( new PieSlice( da[i], dpha[i], i ) );
			}
		}

		double startAngle = 0;
		if ( dTotal == 0 )
		{
			dTotal = 1;
		}

		PieSlice slice = null;
		for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
		{
			slice = (PieSlice) iter.next( );
			double length = ( Math.abs( slice.getPrimitiveValue( ) ) / dTotal ) * 360d;
			double percentage = ( slice.getPrimitiveValue( ) / dTotal ) * 100d;
			slice.setStartAngle( startAngle );
			slice.setSliceLength( length );
			slice.setPercentage( percentage );
			startAngle += length;
		}

		// complement the last slice with residual angle
		// TODO What is this for?
		if ( startAngle > 0 && 360 - startAngle > 0.001 ) // handle precision
		// loss during
		// computations
		{
			slice.setSliceLength( 360 - slice.getStartAngle( ) );
		}

		initExploded( );
	}

	/**
	 * 
	 * @param idr
	 * @throws RenderingException
	 */
	private final void renderDataPoints( IDeviceRenderer idr )
			throws ChartException
	{
		final ScriptHandler sh = pie.getRunTimeContext( ).getScriptHandler( );
		int iTextRenderType = TextRenderEvent.RENDER_TEXT_IN_BLOCK;
		if ( lpDataPoint.getValue( ) == Position.OUTSIDE )
		{
			// RENDER SHADOWS (IF ANY) FIRST
			if ( laDataPoint.getShadowColor( ) != null )
			{
				for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
				{
					PieSlice slice = (PieSlice) iter.next( );
					slice.renderLabel( idr,
							TextRenderEvent.RENDER_SHADOW_AT_LOCATION );
				}
			}
			iTextRenderType = TextRenderEvent.RENDER_TEXT_AT_LOCATION;
		}

		// RENDER ACTUAL TEXT CAPTIONS ON DATA POINTS
		for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
		{
			PieSlice slice = (PieSlice) iter.next( );
			laDataPoint = LabelImpl.copyInstance( ps.getLabel( ) );
			laDataPoint.getCaption( ).setValue( slice.getDisplayValue( ) );
			ScriptHandler.callFunction( sh,
					ScriptHandler.BEFORE_DRAW_DATA_POINT_LABEL,
					slice.getDataPointHints( ),
					laDataPoint,
					pie.getRunTimeContext( ).getScriptContext( ) );
			pie.getRunTimeContext( )
					.notifyStructureChange( IStructureDefinitionListener.BEFORE_DRAW_DATA_POINT_LABEL,
							laDataPoint );

			if ( laDataPoint.isVisible( ) )
			{
				slice.renderLabel( idr, iTextRenderType );
			}
			ScriptHandler.callFunction( sh,
					ScriptHandler.AFTER_DRAW_DATA_POINT_LABEL,
					slice.getDataPointHints( ),
					laDataPoint,
					pie.getRunTimeContext( ).getScriptContext( ) );
			pie.getRunTimeContext( )
					.notifyStructureChange( IStructureDefinitionListener.AFTER_DRAW_DATA_POINT_LABEL,
							laDataPoint );
		}
	}

	/**
	 * 
	 * @param bo
	 */
	/**
	 * 
	 * @param bo
	 */
	private final void computeLabelBounds( Bounds bo, boolean isOutside )
			throws IllegalArgumentException
	{
		// compute additional bottom tick size due to series thickness.

		for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
		{
			PieSlice slice = (PieSlice) iter.next( );
			slice.setBounds( bo );
			if ( isOutside )
			{
				slice.computeLabelBoundOutside( lls, dLeaderLength );
			}
			else
			{
				slice.computeLabelBoundInside( );
			}
		}
	}

	private int getOverlapCount( PieSlice pieSlice )
	{
		int count = 0;
		if ( pieSlice.getLabelBounding( ) != null )
		{
			for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
			{
				PieSlice slice = (PieSlice) iter.next( );
				if ( pieSlice.isLabelOverlap( slice ) )
				{
					count++;
				}
			}
		}
		return count;
	}

	private List getOverlapSliceList( )
	{
		ArrayList list = new ArrayList( );
		for ( int i = 0; i < pieSliceList.size( ); i++ )
		{
			PieSlice slice = (PieSlice) pieSliceList.get( i );
			for ( int j = i + 1; j < pieSliceList.size( ); j++ )
			{
				if ( slice.isLabelOverlap( (PieSlice) pieSliceList.get( j ) ) )
				{
					list.add( slice );
					break;
				}
			}
		}
		if ( !list.isEmpty( ) )
		{
			Collections.sort( list, new Comparator( ) {

				public int compare( Object arg0, Object arg1 )
				{
					PieSlice slice0 = (PieSlice) arg0;
					PieSlice slice1 = (PieSlice) arg1;
					return ( getOverlapCount( slice0 ) - getOverlapCount( slice1 ) )
							* 1000
							+ (int) ( slice1.getdMidAngle( ) - slice0.getdMidAngle( ) );
				}
			} );
		}
		return list;
	}

	private boolean isOverlap( PieSlice pieSlice )
	{
		if ( pieSlice.getLabelBounding( ) != null )
		{
			for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
			{
				PieSlice slice = (PieSlice) iter.next( );
				if ( pieSlice.isLabelOverlap( slice ) )
				{
					if ( pieSlice.getSliceLength( ) == 0
							&& slice.getSliceLength( ) == 0 )
					{//Added to resolve empty category stacking (Bugzilla 197128)
						pieSlice.removeLabelBounding( );
						return false;
					}
					return true;
				}
			}
		}
		return false;
	}

	private void resolveOverlap( PieSlice slice, Bounds bo,
			LeaderLineStyle lls, double dLeaderLength )
	{
		if ( !isOverlap( slice ) )
		{
			return;
		}
		// Clone for backup
		PieSlice clonedSlice = (PieSlice) slice.clone( );
		if ( lls == LeaderLineStyle.STRETCH_TO_SIDE_LITERAL )
		{
			dLeaderLength = LENGTH_HINTS / 2;
			slice.computeLabelBoundOutside( LeaderLineStyle.FIXED_LENGTH_LITERAL,
					dLeaderLength );
			if ( !isOverlap( slice ) && !slice.isLabelClipped( bo ) )
			{
				return;
			}
		}
		double maxOffset = Math.min( dLeaderLength, LENGTH_HINTS
				- dLeaderLength );
		for ( double offset = 1; offset <= maxOffset; offset++ )
		{
			for ( int i = -1; i <= 1; i += 2 )
			{
				double dLength = dLeaderLength + i * offset;
				slice.computeLabelBoundOutside( LeaderLineStyle.FIXED_LENGTH_LITERAL,
						dLength );
				if ( !isOverlap( slice ) && !slice.isLabelClipped( bo ) )
				{
					return;
				}
			}
		}
		if ( LeaderLineStyle.FIXED_LENGTH_LITERAL.equals( lls ) )
		{
			slice.computeLabelBoundOutside( LeaderLineStyle.STRETCH_TO_SIDE_LITERAL,
					dLeaderLength );
			if ( !isOverlap( slice ) && !slice.isLabelClipped( bo ) )
			{
				return;
			}
		}
		double dStart = 0, dEnd = LENGTH_HINTS;
		if ( dLeaderLength > LENGTH_HINTS / 2 )
		{
			dEnd = dLeaderLength - maxOffset;
		}
		else
		{
			dStart = dLeaderLength + maxOffset;
		}
		for ( double l = dStart; l <= dEnd; l++ )
		{
			slice.computeLabelBoundOutside( LeaderLineStyle.FIXED_LENGTH_LITERAL,
					l );
			if ( !isOverlap( slice ) && !slice.isLabelClipped( bo ) )
			{
				return;
			}
		}
		// No solution,restore data
		clonedSlice.copyTo( slice );
	}

	/**
	 * 
	 * @param bo
	 * @param boAdjusted
	 * @param ins
	 * @return
	 * @throws UnexpectedInputException
	 */
	private final Insets adjust( Bounds bo, Bounds boAdjusted, Insets ins )
			throws IllegalArgumentException
	{
		computeLabelBounds( boAdjusted, true );
		ins.set( 0, 0, 0, 0 );
		double dDelta = 0;
		for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
		{
			PieSlice slice = (PieSlice) iter.next( );
			BoundingBox bb = slice.getLabelBounding( );
			if ( bb.getLeft( ) < bo.getLeft( ) )
			{
				dDelta = bo.getLeft( ) - bb.getLeft( );
				if ( ins.getLeft( ) < dDelta )
				{
					ins.setLeft( dDelta );
				}
			}
			if ( bb.getTop( ) < bo.getTop( ) )
			{
				dDelta = bo.getTop( ) - bb.getTop( );
				if ( ins.getTop( ) < dDelta )
				{
					ins.setTop( dDelta );
				}
			}
			if ( bb.getLeft( ) + bb.getWidth( ) > bo.getLeft( ) + bo.getWidth( ) )
			{
				dDelta = bb.getLeft( )
						+ bb.getWidth( )
						- bo.getLeft( )
						- bo.getWidth( );
				if ( ins.getRight( ) < dDelta )
				{
					ins.setRight( dDelta );
				}
			}
			if ( bb.getTop( ) + bb.getHeight( ) > bo.getTop( ) + bo.getHeight( ) )
			{
				dDelta = bb.getTop( )
						+ bb.getHeight( )
						- bo.getTop( )
						- bo.getHeight( );
				if ( ins.getBottom( ) < dDelta )
				{
					ins.setBottom( dDelta );
				}
			}
		}
		return ins;
	}

	/**
	 * 
	 * @param bo
	 * @throws UndefinedValueException
	 * @throws UnexpectedInputException
	 */
	final void computeInsets( Bounds bo ) throws ChartException,
			IllegalArgumentException
	{
		boSetDuringComputation = BoundsImpl.copyInstance( bo );
		xs = pie.getXServer( );

		// ALLOCATE SPACE FOR THE SERIES TITLE
		boTitleContainer = null;
		if ( laSeriesTitle.isVisible( ) )
		{
			if ( lpSeriesTitle == null )
			{
				throw new ChartException( ChartEngineExtensionPlugin.ID,
						ChartException.UNDEFINED_VALUE,
						"exception.unspecified.visible.series.title", //$NON-NLS-1$
						Messages.getResourceBundle( pie.getRunTimeContext( )
								.getULocale( ) ) );
			}

			final BoundingBox bb = Methods.computeBox( xs,
					IConstants.BELOW,
					laSeriesTitle,
					0,
					0 );
			boTitleContainer = BoundsImpl.create( 0, 0, 0, 0 );

			switch ( lpSeriesTitle.getValue( ) )
			{
				case Position.BELOW :
					bo.setHeight( bo.getHeight( ) - bb.getHeight( ) );
					boTitleContainer.set( bo.getLeft( ), bo.getTop( )
							+ bo.getHeight( ), bo.getWidth( ), bb.getHeight( ) );
					break;
				case Position.ABOVE :
					boTitleContainer.set( bo.getLeft( ),
							bo.getTop( ),
							bo.getWidth( ),
							bb.getHeight( ) );
					bo.setTop( bo.getTop( ) + bb.getHeight( ) );
					bo.setHeight( bo.getHeight( ) - bb.getHeight( ) );
					break;
				case Position.LEFT :
					bo.setWidth( bo.getWidth( ) - bb.getWidth( ) );
					boTitleContainer.set( bo.getLeft( ),
							bo.getTop( ),
							bb.getWidth( ),
							bo.getHeight( ) );
					bo.setLeft( bo.getLeft( ) + bb.getWidth( ) );
					break;
				case Position.RIGHT :
					bo.setWidth( bo.getWidth( ) - bb.getWidth( ) );
					boTitleContainer.set( bo.getLeft( ) + bo.getWidth( ),
							bo.getTop( ),
							bb.getWidth( ),
							bo.getHeight( ) );
					break;
				default :
					throw new IllegalArgumentException( MessageFormat.format( Messages.getResourceBundle( pie.getRunTimeContext( )
							.getULocale( ) )
							.getString( "exception.illegal.pie.series.title.position" ), //$NON-NLS-1$
							new Object[]{
								lpSeriesTitle
							} )

					);
			}
		}

		if ( !laDataPoint.isSetVisible( ) )
		{
			throw new ChartException( ChartEngineExtensionPlugin.ID,
					ChartException.UNDEFINED_VALUE,
					"exception.unspecified.datapoint.visibility.pie", //$NON-NLS-1$
					new Object[]{
						ps
					},
					Messages.getResourceBundle( pie.getRunTimeContext( )
							.getULocale( ) ) );
		}

		if ( lpDataPoint == Position.OUTSIDE_LITERAL )
		{
			if ( laDataPoint.isVisible( ) ) // FILTERED FOR PERFORMANCE GAIN
			{
				// ADJUST THE BOUNDS TO ACCOMODATE THE DATA POINT LABELS +
				// LEADER LINES RENDERED OUTSIDE
				Bounds boBeforeAdjusted = BoundsImpl.copyInstance( bo );
				Bounds boAdjusted = BoundsImpl.copyInstance( bo );
				Insets insTrim = InsetsImpl.create( 0, 0, 0, 0 );
				do
				{
					adjust( bo, boAdjusted, insTrim );
					boAdjusted.adjust( insTrim );
				} while ( !insTrim.areLessThan( 0.5 )
						&& boAdjusted.getWidth( ) > 0
						&& boAdjusted.getHeight( ) > 0 );
				bo = boAdjusted;
			}
		}
		else if ( lpDataPoint == Position.INSIDE_LITERAL )
		{
			if ( laDataPoint.isVisible( ) ) // FILTERED FOR PERFORMANCE GAIN
			{
				computeLabelBounds( bo, false );
			}
		}
		else
		{
			throw new IllegalArgumentException( MessageFormat.format( Messages.getResourceBundle( pie.getRunTimeContext( )
					.getULocale( ) )
					.getString( "exception.invalid.datapoint.position.pie" ), //$NON-NLS-1$
					new Object[]{
						lpDataPoint
					} )

			);
		}

		insCA = InsetsImpl.create( bo.getTop( )
				- boSetDuringComputation.getTop( ),
				bo.getLeft( ) - boSetDuringComputation.getLeft( ),
				boSetDuringComputation.getTop( )
						+ boSetDuringComputation.getHeight( )
						- ( bo.getTop( ) + bo.getHeight( ) ),
				boSetDuringComputation.getLeft( )
						+ boSetDuringComputation.getWidth( )
						- ( bo.getLeft( ) + bo.getWidth( ) ) );
		bBoundsAdjustedForInsets = false;
	}

	/**
	 * 
	 * @return
	 */
	final Insets getFittingInsets( )
	{
		return insCA;
	}

	/**
	 * 
	 * @param insCA
	 */
	final void setFittingInsets( Insets insCA ) throws IllegalArgumentException
	{
		this.insCA = insCA;
		if ( !bBoundsAdjustedForInsets ) // CHECK IF PREVIOUSLY ADJUSTED
		{
			bBoundsAdjustedForInsets = true;
			boSetDuringComputation.adjust( insCA );
		}

		if ( lpDataPoint == Position.OUTSIDE_LITERAL )
		{
			if ( laDataPoint.isVisible( ) ) // FILTERED FOR PERFORMANCE GAIN
			{
				computeLabelBounds( boSetDuringComputation, true );
			}
		}
		else if ( lpDataPoint == Position.INSIDE_LITERAL )
		{
			if ( laDataPoint.isVisible( ) ) // FILTERED FOR PERFORMANCE GAIN
			{
				computeLabelBounds( boSetDuringComputation, false );
			}
		}
	}

	/**
	 * 
	 * @param idr
	 * @param bo
	 * @throws RenderingException
	 */
	public final void render( IDeviceRenderer idr, Bounds bo )
			throws ChartException
	{
		Bounds boBeforeAdjusted = BoundsImpl.copyInstance( bo );
		bo.adjust( insCA );

		xs = idr.getDisplayServer( );
		this.idr = idr;
		if ( !laSeriesTitle.isSetVisible( ) )
		{
			throw new ChartException( ChartEngineExtensionPlugin.ID,
					ChartException.RENDERING,
					"exception.unspecified.pie.series.title.visibility",//$NON-NLS-1$
					new Object[]{
						ps
					},
					Messages.getResourceBundle( pie.getRunTimeContext( )
							.getULocale( ) ) );
		}

		final ScriptHandler sh = pie.getRunTimeContext( ).getScriptHandler( );

		double w = bo.getWidth( ) / 2d - dExplosion;
		double h = bo.getHeight( ) / 2d - dExplosion - dThickness / 2d;

		double xc = bo.getLeft( ) + bo.getWidth( ) / 2d;
		double yc = bo.getTop( ) + bo.getHeight( ) / 2d;

		if ( ratio > 0 && w > 0 )
		{
			if ( h / w > ratio )
			{
				h = w * ratio;
			}
			else if ( h / w < ratio )
			{
				w = h / ratio;
			}
		}

		// detect invalid rendering size.
		if ( w > 0 && h > 0 )
		{
			// RENDER THE INSIDE OF THE PIE SLICES AS DEFERRED PLANES (FLAT AND
			// CURVED)
			if ( dThickness > 0 )
			{
				for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
				{
					PieSlice slice = (PieSlice) iter.next( );
					Fill fPaletteEntry = null;
					if ( bPaletteByCategory )
					{
						if ( bMinSliceApplied
								&& pieSliceList.indexOf( slice ) == pieSliceList.size( ) - 1 )
						{
							fPaletteEntry = getPaletteColor( orginalSliceCount );
						}
						else
						{
							fPaletteEntry = getPaletteColor( slice.getCategoryIndex( ) );
						}
					}
					else
					{
						fPaletteEntry = getPaletteColor( pie.getSeriesDefinition( )
								.getRunTimeSeries( )
								.indexOf( ps ) );
					}
					ScriptHandler.callFunction( sh,
							ScriptHandler.BEFORE_DRAW_ELEMENT,
							slice.getDataPointHints( ),
							fPaletteEntry );
					ScriptHandler.callFunction( sh,
							ScriptHandler.BEFORE_DRAW_DATA_POINT,
							slice.getDataPointHints( ),
							fPaletteEntry,
							pie.getRunTimeContext( ).getScriptContext( ) );
					pie.getRunTimeContext( )
							.notifyStructureChange( IStructureDefinitionListener.BEFORE_DRAW_ELEMENT,
									slice.getDataPointHints( ) );
					pie.getRunTimeContext( )
							.notifyStructureChange( IStructureDefinitionListener.BEFORE_DRAW_DATA_POINT,
									slice.getDataPointHints( ) );
					slice.render( LocationImpl.create( xc, yc ),
							LocationImpl.create( 0, dThickness ),
							SizeImpl.create( w, h ),
							fPaletteEntry,
							LOWER );
					ScriptHandler.callFunction( sh,
							ScriptHandler.AFTER_DRAW_ELEMENT,
							slice.getDataPointHints( ),
							fPaletteEntry );
					ScriptHandler.callFunction( sh,
							ScriptHandler.AFTER_DRAW_DATA_POINT,
							slice.getDataPointHints( ),
							fPaletteEntry,
							pie.getRunTimeContext( ).getScriptContext( ) );
					pie.getRunTimeContext( )
							.notifyStructureChange( IStructureDefinitionListener.AFTER_DRAW_ELEMENT,
									slice.getDataPointHints( ) );
					pie.getRunTimeContext( )
							.notifyStructureChange( IStructureDefinitionListener.AFTER_DRAW_DATA_POINT,
									slice.getDataPointHints( ) );
				}

				// SORT AND ACTUALLY RENDER THE PLANES PREVIOUSLY WRITTEN AS
				// DEFERRED
				sortAndRenderPlanes( );
			}

			// RENDER THE UPPER SECTORS ON THE PIE SLICES (DON'T CARE ABOUT
			// THE ORDER)
			for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
			{
				PieSlice slice = (PieSlice) iter.next( );
				Fill fPaletteEntry = null;
				if ( bPaletteByCategory )
				{
					if ( bMinSliceApplied
							&& pieSliceList.indexOf( slice ) == pieSliceList.size( ) - 1 )
					{
						fPaletteEntry = getPaletteColor( orginalSliceCount );
					}
					else
					{
						fPaletteEntry = getPaletteColor( slice.getCategoryIndex( ) );
					}
				}
				else
				{
					fPaletteEntry = getPaletteColor( pie.getSeriesDefinition( )
							.getRunTimeSeries( )
							.indexOf( ps ) );
				}
				ScriptHandler.callFunction( sh,
						ScriptHandler.BEFORE_DRAW_DATA_POINT,
						slice.getDataPointHints( ),
						fPaletteEntry,
						pie.getRunTimeContext( ).getScriptContext( ) );
				pie.getRunTimeContext( )
						.notifyStructureChange( IStructureDefinitionListener.BEFORE_DRAW_DATA_POINT,
								slice.getDataPointHints( ) );
				slice.render( LocationImpl.create( xc, yc ),
						LocationImpl.create( 0, dThickness ),
						SizeImpl.create( w, h ),
						fPaletteEntry,
						UPPER );
				ScriptHandler.callFunction( sh,
						ScriptHandler.AFTER_DRAW_DATA_POINT,
						slice.getDataPointHints( ),
						fPaletteEntry,
						pie.getRunTimeContext( ).getScriptContext( ) );
				pie.getRunTimeContext( )
						.notifyStructureChange( IStructureDefinitionListener.AFTER_DRAW_DATA_POINT,
								slice.getDataPointHints( ) );
			}

		}

		// RENDER THE SERIES TITLE NOW
		ScriptHandler.callFunction( sh,
				ScriptHandler.BEFORE_DRAW_SERIES_TITLE,
				ps,
				laSeriesTitle,
				pie.getRunTimeContext( ).getScriptContext( ) );
		pie.getRunTimeContext( )
				.notifyStructureChange( IStructureDefinitionListener.BEFORE_DRAW_SERIES_TITLE,
						laSeriesTitle );
		laDataPoint = LabelImpl.copyInstance( ps.getLabel( ) );
		if ( laSeriesTitle.isVisible( ) )
		{
			final TextRenderEvent tre = (TextRenderEvent) ( (EventObjectCache) idr ).getEventObject( WrappedStructureSource.createSeriesTitle( ps,
					laSeriesTitle ),
					TextRenderEvent.class );
			tre.setLabel( laSeriesTitle );
			tre.setBlockBounds( boTitleContainer );
			tre.setBlockAlignment( null );
			tre.setAction( TextRenderEvent.RENDER_TEXT_IN_BLOCK );
			idr.drawText( tre );
		}
		ScriptHandler.callFunction( sh,
				ScriptHandler.AFTER_DRAW_SERIES_TITLE,
				ps,
				laSeriesTitle,
				pie.getRunTimeContext( ).getScriptContext( ) );
		pie.getRunTimeContext( )
				.notifyStructureChange( IStructureDefinitionListener.AFTER_DRAW_SERIES_TITLE,
						laSeriesTitle );

		// LASTLY, RENDER THE DATA POINTS
		if ( laDataPoint.isVisible( ) )
		{ // IF NOT VISIBLE, DON'T DO ANYTHING
			try
			{
				if ( laDataPoint.getCaption( ).getFont( ).getRotation( ) == 0 )
				{
					// Currently do not support rotated label
					for ( Iterator iter = getOverlapSliceList( ).iterator( ); iter.hasNext( ); )
					{
						PieSlice slice = (PieSlice) iter.next( );
						resolveOverlap( slice,
								boBeforeAdjusted,
								lls,
								dLeaderLength );
					}
				}
				renderDataPoints( idr );
			}
			catch ( ChartException rex )
			{
				logger.log( rex );
			}
		}
	}

	/**
	 * 
	 * @param areBentOrTwistedCurve
	 * @param dX1
	 * @param dX2
	 */
	private final void deferCurvedPlane( AreaRenderEvent areBentOrTwistedCurve,
			double dX1, double dX2 )
	{
		alPlanes.add( new CurvedPlane( areBentOrTwistedCurve, dX1, dX2 ) );
	}

	/**
	 * Defer to draw curved outline.
	 * 
	 * @param areBentOrTwistedCurve
	 *            outline group.
	 */
	private final void deferCurvedOutline( AreaRenderEvent areBentOrTwistedCurve )
	{
		alPlanes.add( new CurvedPlane( areBentOrTwistedCurve ) );
	}

	/**
	 * 
	 * @param daXPoints
	 * @param daYPoints
	 * @param cd
	 * @param dX1
	 * @param dX2
	 */
	private final void deferFlatPlane( double[] daXPoints, double[] daYPoints,
			Fill cd, double dX1, double dX2, DataPointHints dph )
	{
		alPlanes.add( new FlatPlane( daXPoints, daYPoints, cd, dX1, dX2, dph ) );
	}

	/**
	 * 
	 * @throws RenderingException
	 */
	private final void sortAndRenderPlanes( ) throws ChartException
	{
		int n = alPlanes.size( );
		IDrawable id;
		Collections.sort( alPlanes );
		for ( int i = 0; i < n; i++ )
		{
			id = (IDrawable) alPlanes.get( i );
			id.draw( );
		}

		alPlanes.clear( );
	}

	private final ColorDefinition getSliceOutline( Fill f )
	{
		if ( ps.getSliceOutline( ) == null )
		{
			if ( f instanceof ColorDefinition )
			{
				return ( (ColorDefinition) f ).darker( );
			}
			else
			{
				return ColorDefinitionImpl.TRANSPARENT( );
			}
		}
		return ColorDefinitionImpl.copyInstance( ps.getSliceOutline( ) );
	}

	private void initExploded( )
	{

		if ( sExplosionExpression == null
				|| !pie.getRunTimeContext( ).isScriptingEnabled( ) )
		{
			return;
		}

		for ( Iterator iter = pieSliceList.iterator( ); iter.hasNext( ); )
		{
			PieSlice slice = (PieSlice) iter.next( );
			try
			{
				pie.getRunTimeContext( )
						.getScriptHandler( )
						.registerVariable( ScriptHandler.BASE_VALUE,
								slice.getDataPointHints( ).getBaseValue( ) );
				pie.getRunTimeContext( )
						.getScriptHandler( )
						.registerVariable( ScriptHandler.ORTHOGONAL_VALUE,
								slice.getDataPointHints( ).getOrthogonalValue( ) );
				pie.getRunTimeContext( )
						.getScriptHandler( )
						.registerVariable( ScriptHandler.SERIES_VALUE,
								slice.getDataPointHints( ).getSeriesValue( ) );

				Object obj = pie.getRunTimeContext( )
						.getScriptHandler( )
						.evaluate( sExplosionExpression );

				if ( obj instanceof Boolean )
				{
					slice.setExploded( ( (Boolean) obj ).booleanValue( ) );
				}

				pie.getRunTimeContext( )
						.getScriptHandler( )
						.unregisterVariable( ScriptHandler.BASE_VALUE );
				pie.getRunTimeContext( )
						.getScriptHandler( )
						.unregisterVariable( ScriptHandler.ORTHOGONAL_VALUE );
				pie.getRunTimeContext( )
						.getScriptHandler( )
						.unregisterVariable( ScriptHandler.SERIES_VALUE );

			}
			catch ( ChartException e )
			{
				logger.log( e );
			}
		}
	}

	// HANDLE ELEVATION COMPUTATION FOR SOLID
	// COLORS,
	// GRADIENTS AND IMAGES

	protected Gradient getDepthGradient( Fill cd )
	{
		if ( cd instanceof Gradient )
		{
			return GradientImpl.create( ( (Gradient) cd ).getStartColor( )
					.darker( ),
					( (Gradient) cd ).getEndColor( ).darker( ),
					( (Gradient) cd ).getDirection( ),
					( (Gradient) cd ).isCyclic( ) );

		}
		else
			return GradientImpl.create( ( cd instanceof ColorDefinition ) ? ( (ColorDefinition) cd ).darker( )
					: ColorDefinitionImpl.GREY( ),
					ColorDefinitionImpl.BLACK( ),
					0,
					true );
	}

	/**
	 * @param cd
	 * @param startAngle
	 * @param endAngle
	 * @return
	 */
	protected Gradient getDepthGradient( Fill cd, double startAngle,
			double endAngle )
	{
		if ( cd instanceof Gradient )
		{
			return GradientImpl.create( ( (Gradient) cd ).getStartColor( )
					.darker( ),
					( (Gradient) cd ).getEndColor( ).darker( ),
					( (Gradient) cd ).getDirection( ),
					( (Gradient) cd ).isCyclic( ) );

		}
		else
		{
			ColorDefinition standCD = ( cd instanceof ColorDefinition ) ? ( (ColorDefinition) cd ).darker( )
					: ColorDefinitionImpl.GREY( );
			float[] hsbvals = Color.RGBtoHSB( standCD.getRed( ),
					standCD.getGreen( ),
					standCD.getBlue( ),
					null );
			float[] startHSB = new float[3];
			startHSB[0] = hsbvals[0];
			startHSB[1] = hsbvals[1];
			startHSB[2] = hsbvals[2];

			float[] endHSB = new float[3];
			endHSB[0] = hsbvals[0];
			endHSB[1] = hsbvals[1];
			endHSB[2] = hsbvals[2];

			float brightAlpha = 1f / 180;
			if ( startAngle < 180 )
			{
				startHSB[2] = (float) ( startAngle * brightAlpha );
			}
			else
			{
				startHSB[2] = (float) ( 1 - ( startAngle - 180 ) * brightAlpha );
			}

			if ( endAngle < 180 )
			{
				endHSB[2] = (float) ( endAngle * brightAlpha );
			}
			else
			{
				endHSB[2] = (float) ( 1 - ( endAngle - 180 ) * brightAlpha );
			}

			Color startColor = new Color( Color.HSBtoRGB( startHSB[0],
					startHSB[1],
					startHSB[2] ) );
			Color endColor = new Color( Color.HSBtoRGB( endHSB[0],
					endHSB[1],
					endHSB[2] ) );

			ColorDefinition startCD = ColorDefinitionImpl.copyInstance( standCD );
			startCD.set( startColor.getRed( ),
					startColor.getGreen( ),
					startColor.getBlue( ) );

			ColorDefinition endCD = ColorDefinitionImpl.copyInstance( standCD );
			endCD.set( endColor.getRed( ),
					endColor.getGreen( ),
					endColor.getBlue( ) );

			if ( endAngle <= 180 )
			{
				return GradientImpl.create( endCD, startCD, 0, true );
			}
			else
			{
				return GradientImpl.create( startCD, endCD, 0, true );
			}
		}
	}

	/**
	 * Used for deferred rendering
	 * 
	 * @param topBound
	 *            The top round bounds of pie slice.
	 * @param bottomBound
	 *            The bottom round bounds of pie slice.
	 * @param dStartAngle
	 *            The start agnle of pie slice.
	 * @param dAngleExtent
	 *            The extent angle of pie slice.
	 * @param lreStartB2T
	 *            The bottom to top line rendering event in start location of
	 *            pie slice.
	 * @param lreEndB2T
	 *            The top to bottom line rendering event in end location of pie
	 *            slice.
	 * @param cd
	 * @param dph
	 *            data point hints.
	 * @param loC
	 *            center point of bottom cycle.
	 * @param loCTop
	 *            center point of top cycle.
	 * @param sz
	 *            width and height of cycle.
	 */
	private final void registerCurvedSurface( Bounds topBound,
			Bounds bottomBound, double dStartAngle, double dAngleExtent,
			LineRenderEvent lreStartB2T, LineRenderEvent lreEndB2T, Fill cd,
			DataPointHints dph, Location loC, Location loCTop, Size sz )
	{
		// 1. Get all splited angles.
		double[] anglePoints = new double[4];

		double endAngle = dStartAngle + dAngleExtent;

		int i = 0;
		anglePoints[i++] = dStartAngle;
		if ( endAngle > 180 && dStartAngle < 180 )
		{
			anglePoints[i++] = 180;
		}
		if ( endAngle > 360 && dStartAngle < 360 )
		{
			anglePoints[i++] = 360;
		}
		anglePoints[i] = endAngle;

		// 2. Do Render.
		if ( i == 1 ) // The simple case, only has one curved plane.
		{
			// CURVED PLANE 1
			final ArcRenderEvent arcRE1 = new ArcRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
					dph ) );
			final ArcRenderEvent arcRE2 = new ArcRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
					dph ) );
			arcRE1.setBounds( topBound );
			arcRE1.setStartAngle( dStartAngle );
			arcRE1.setAngleExtent( dAngleExtent );
			arcRE1.setStyle( ArcRenderEvent.OPEN );
			arcRE2.setBounds( bottomBound );
			arcRE2.setStartAngle( dStartAngle + dAngleExtent );
			arcRE2.setAngleExtent( -dAngleExtent );
			arcRE2.setStyle( ArcRenderEvent.OPEN );
			AreaRenderEvent areRE = new AreaRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
					dph ) );
			areRE.add( lreStartB2T );
			areRE.add( arcRE1 );
			areRE.add( lreEndB2T );
			areRE.add( arcRE2 );

			areRE.setOutline( LineAttributesImpl.create( getSliceOutline( cd ),
					LineStyle.SOLID_LITERAL,
					1 ) );
			areRE.setBackground( getDepthGradient( cd, dStartAngle, dStartAngle
					+ dAngleExtent ) );

			deferCurvedPlane( areRE,
					lreStartB2T.getStart( ).getX( ),
					lreEndB2T.getStart( ).getX( ) );
		}
		else
		// The multiple case, should be more curved plane.
		{
			Stack lineStack = new Stack( );
			AreaRenderEvent areLine = new AreaRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
					dph ) );
			areLine.setOutline( LineAttributesImpl.create( getSliceOutline( cd ),
					LineStyle.SOLID_LITERAL,
					1 ) );
			areLine.setBackground( null );

			for ( int j = 0; j < i; j++ )
			{
				double startAngle = anglePoints[j];
				double angleExtent = anglePoints[j + 1] - anglePoints[j];

				Object[] edgeLines = getEdgeLines( startAngle,
						angleExtent,
						loC,
						loCTop,
						sz,
						dph );

				// CURVED PLANE 1
				final ArcRenderEvent arcRE1 = new ArcRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dph ) );
				final ArcRenderEvent arcRE2 = new ArcRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dph ) );

				arcRE1.setBounds( topBound );
				arcRE1.setStartAngle( startAngle );
				arcRE1.setAngleExtent( angleExtent );
				arcRE1.setStyle( ArcRenderEvent.OPEN );

				arcRE2.setBounds( bottomBound );
				arcRE2.setStartAngle( anglePoints[j + 1] );
				arcRE2.setAngleExtent( -angleExtent );
				arcRE2.setStyle( ArcRenderEvent.OPEN );

				// Fill pie slice.
				AreaRenderEvent are = new AreaRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dph ) );
				are.add( (LineRenderEvent) edgeLines[0] );
				are.add( arcRE1 );
				are.add( (LineRenderEvent) edgeLines[1] );
				are.add( arcRE2 );

				are.setOutline( null );

				are.setBackground( getDepthGradient( cd,
						anglePoints[j],
						anglePoints[j + 1] ) );
				deferCurvedPlane( are,
						( (LineRenderEvent) edgeLines[0] ).getStart( ).getX( ),
						( (LineRenderEvent) edgeLines[1] ).getStart( ).getX( ) );

				// Arrange pie slice outline.
				if ( j == 0 ) // It is first sector of pie slice.
				{
					areLine.add( (LineRenderEvent) edgeLines[0] );
					areLine.add( arcRE1 );
					lineStack.push( arcRE2 );
				}
				else if ( j == ( i - 1 ) ) // It is last sector of pie slice.
				{
					areLine.add( arcRE1 );
					areLine.add( (LineRenderEvent) edgeLines[1] );
					areLine.add( arcRE2 );
				}
				else
				{
					areLine.add( arcRE1 );
					lineStack.push( arcRE2 );
				}
			}

			// Arrange pie slice outline.
			while ( !lineStack.empty( ) )
			{
				areLine.add( (PrimitiveRenderEvent) lineStack.pop( ) );
			}

			// Draw pie slice outline.
			deferCurvedOutline( areLine );
		}
	}

	/**
	 * @param startAngle
	 * @param extentAngle
	 * @param loC
	 * @param loCTop
	 * @param sz
	 * @param dph
	 * @return
	 */
	private final Object[] getEdgeLines( double startAngle, double extentAngle,
			Location loC, Location loCTop, Size sz, DataPointHints dph )
	{

		double dAngleInRadians = Math.toRadians( startAngle );
		double dSineThetaStart = Math.sin( dAngleInRadians );
		double dCosThetaStart = Math.cos( dAngleInRadians );
		dAngleInRadians = Math.toRadians( startAngle + extentAngle );
		double dSineThetaEnd = Math.sin( dAngleInRadians );
		double dCosThetaEnd = Math.cos( dAngleInRadians );

		double xE = ( sz.getWidth( ) * dCosThetaEnd );
		double yE = ( sz.getHeight( ) * dSineThetaEnd );
		double xS = ( sz.getWidth( ) * dCosThetaStart );
		double yS = ( sz.getHeight( ) * dSineThetaStart );

		final LineRenderEvent lreStartB2T = new LineRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
				dph ) );
		lreStartB2T.setStart( LocationImpl.create( loC.getX( ) + xS, loC.getY( )
				- yS ) );
		lreStartB2T.setEnd( LocationImpl.create( loCTop.getX( ) + xS,
				loCTop.getY( ) - yS ) );
		final LineRenderEvent lreEndT2B = new LineRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
				dph ) );
		lreEndT2B.setStart( LocationImpl.create( loCTop.getX( ) + xE,
				loCTop.getY( ) - yE ) );
		lreEndT2B.setEnd( LocationImpl.create( loC.getX( ) + xE, loC.getY( )
				- yE ) );

		return new Object[]{
				lreStartB2T, lreEndT2B
		};
	}

	/**
	 * 
	 * @param iIndex
	 * @return
	 */
	private final Fill getPaletteColor( int iIndex )
	{
		final Fill fiClone = (Fill) EcoreUtil.copy( (Fill) pa.getEntries( )
				.get( iIndex % pa.getEntries( ).size( ) ) );

		pie.updateTranslucency( fiClone, ps );
		return fiClone;
	}

	/**
	 * CurvedPlane
	 */
	private final class CurvedPlane implements Comparable, IDrawable
	{

		private final AreaRenderEvent _are;

		private final Bounds _bo;

		CurvedPlane( AreaRenderEvent are, double dX1, double dX2 )
		{
			_are = are;
			_bo = are.getBounds( );
		}

		/**
		 * Constructor of the class.
		 * 
		 * 
		 */
		CurvedPlane( AreaRenderEvent are )
		{
			this( are, 0d, 0d );
		}

		public final Bounds getBounds( )
		{
			return _bo;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public final int compareTo( Object o ) // Z-ORDER TEST
		{
			final CurvedPlane cp1 = this;
			if ( o instanceof CurvedPlane )
			{
				final CurvedPlane cp2 = (CurvedPlane) o;
				final double dMinY1 = cp1.getMinY( );
				final double dMinY2 = cp2.getMinY( );
				double dDiff = dMinY1 - dMinY2;
				if ( !ChartUtil.mathEqual( dDiff, 0 ) )
				{
					return ( dDiff < 0 ) ? LESS : ( dDiff > 0 ) ? MORE : EQUAL;
				}
				else
				{
					final double dMaxY1 = cp1.getMaxY( );
					final double dMaxY2 = cp2.getMaxY( );
					dDiff = dMaxY1 - dMaxY2;
					if ( !ChartUtil.mathEqual( dDiff, 0 ) )
					{
						return ( dDiff < 0 ) ? LESS : MORE;
					}
					else
					{
						final double dMinX1 = cp1.getMinX( );
						final double dMinX2 = cp2.getMinX( );
						dDiff = dMinX1 - dMinX2;
						if ( !ChartUtil.mathEqual( dDiff, 0 ) )
						{
							return ( dDiff < 0 ) ? LESS : MORE;
						}
						else
						{
							final double dMaxX1 = cp1.getMaxX( );
							final double dMaxX2 = cp2.getMaxX( );
							dDiff = dMaxX1 - dMaxX2;
							if ( !ChartUtil.mathEqual( dDiff, 0 ) )
							{
								return ( dDiff < 0 ) ? LESS : MORE;
							}
							else
							{
								return EQUAL;
							}
						}
					}
				}
			}
			else if ( o instanceof FlatPlane )
			{
				final FlatPlane pi2 = (FlatPlane) o;
				return pi2.compareTo( cp1 ) * -1; // DELEGATE AND INVERT
				// RESULT
			}
			return EQUAL;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.birt.chart.prototype.Pie.IDrawable#draw(java.awt.Graphics2D)
		 */
		public final void draw( ) throws ChartException
		{
			idr.fillArea( _are );
			idr.drawArea( _are );
			/*
			 * GradientPaint gp = new GradientPaint( (float)dGradientStart2D, 0,
			 * Color.black, (float)(dGradientStart2D + (dGradientEnd2D -
			 * dGradientStart2D)/2), 0, _c, true); g2d.setPaint(gp);
			 * g2d.fill(_sh); g2d.setColor(_c.darker()); g2d.draw(_sh);
			 */
		}

		private final double getMinY( )
		{
			return _bo.getTop( );
		}

		private final double getMinX( )
		{
			return _bo.getLeft( );
		}

		private final double getMaxX( )
		{
			return _bo.getLeft( ) + _bo.getWidth( );
		}

		private final double getMaxY( )
		{
			return _bo.getTop( ) + _bo.getHeight( );
		}
	}

	/**
	 * FlatPlane
	 */
	private final class FlatPlane implements Comparable, IDrawable
	{

		private final double[] _daXPoints, _daYPoints;

		private Fill _cd;

		private final Bounds _bo;

		private DataPointHints _dph;

		FlatPlane( double[] daXPoints, double[] daYPoints, Fill cd, double dX1,
				double dX2, DataPointHints dph )
		{
			_daXPoints = daXPoints;
			_daYPoints = daYPoints;
			_dph = dph;

			// COMPUTE THE BOUNDS
			final int n = _daXPoints.length;
			double dMinX = 0, dMinY = 0, dMaxX = 0, dMaxY = 0;

			for ( int i = 0; i < n; i++ )
			{
				if ( i == 0 )
				{
					dMinX = _daXPoints[i];
					dMinY = _daYPoints[i];
					dMaxX = dMinX;
					dMaxY = dMinY;
				}
				else
				{
					if ( dMinX > _daXPoints[i] )
					{
						dMinX = _daXPoints[i];
					}
					if ( dMinY > _daYPoints[i] )
					{
						dMinY = _daYPoints[i];
					}
					if ( dMaxX < _daXPoints[i] )
					{
						dMaxX = _daXPoints[i];
					}
					if ( dMaxY < _daYPoints[i] )
					{
						dMaxY = _daYPoints[i];
					}
				}
			}
			_bo = BoundsImpl.create( dMinX, dMinY, dMaxX - dMinX, dMaxY - dMinY );

			_cd = cd;

			int nPoints = _daXPoints.length;
			int[] iaX = new int[nPoints];
			int[] iaY = new int[nPoints];
			for ( int i = 0; i < nPoints; i++ )
			{
				iaX[i] = (int) daXPoints[i];
				iaY[i] = (int) daYPoints[i];
			}
			// _p = new Polygon(iaX, iaY, nPoints);
		}

		public Bounds getBounds( )
		{
			return _bo;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.birt.chart.prototype.Pie.IDrawable#draw(java.awt.Graphics2D)
		 */
		public final void draw( ) throws ChartException
		{
			PolygonRenderEvent pre = (PolygonRenderEvent) ( (EventObjectCache) idr ).getEventObject( WrappedStructureSource.createSeriesDataPoint( ps,
					_dph ),
					PolygonRenderEvent.class );
			pre.setPoints( toLocationArray( ) );
			liaEdges.setColor( getSliceOutline( _cd ) );
			pre.setOutline( liaEdges );

			pre.setBackground( getDepthGradient( _cd ) );
			idr.fillPolygon( pre );
			idr.drawPolygon( pre );
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public final int compareTo( Object o ) // Z-ORDER TEST
		{
			final FlatPlane pi1 = this;
			if ( o instanceof FlatPlane )
			{
				final FlatPlane pi2 = (FlatPlane) o;

				final double dMinY1 = pi1.getMinY( );
				final double dMinY2 = pi2.getMinY( );
				double dDiff = dMinY1 - dMinY2;
				if ( !ChartUtil.mathEqual( dDiff, 0 ) )
				{
					return ( dDiff < 0 ) ? LESS : ( dDiff > 0 ) ? MORE : EQUAL;
				}
				else
				{
					final double dMaxY1 = pi1.getMaxY( );
					final double dMaxY2 = pi2.getMaxY( );
					dDiff = dMaxY1 - dMaxY2;
					if ( !ChartUtil.mathEqual( dDiff, 0 ) )
					{
						return ( dDiff < 0 ) ? LESS : MORE;
					}
					else
					{
						final double dMinX1 = pi1.getMinX( );
						final double dMinX2 = pi2.getMinX( );
						dDiff = dMinX1 - dMinX2;
						if ( !ChartUtil.mathEqual( dDiff, 0 ) )
						{
							return ( dDiff < 0 ) ? LESS : MORE;
						}
						else
						{
							final double dMaxX1 = pi1.getMaxX( );
							final double dMaxX2 = pi2.getMaxX( );
							dDiff = dMaxX1 - dMaxX2;
							if ( !ChartUtil.mathEqual( dDiff, 0 ) )
							{
								return ( dDiff < 0 ) ? LESS : MORE;
							}
							else
							{
								return EQUAL;
							}
						}
					}
				}
			}
			else if ( o instanceof CurvedPlane )
			{
				final CurvedPlane pi2 = (CurvedPlane) o;

				final double dMinY1 = pi1.getMinY( );
				final double dMinY2 = pi2.getMinY( );
				double dDiff = dMinY1 - dMinY2;
				if ( !ChartUtil.mathEqual( dDiff, 0 ) )
				{
					return ( dDiff < 0 ) ? LESS : MORE;
				}
				else
				{
					final double dMaxY1 = pi1.getMaxY( );
					final double dMaxY2 = pi2.getMaxY( );
					dDiff = dMaxY1 - dMaxY2;
					if ( !ChartUtil.mathEqual( dDiff, 0 ) )
					{
						return ( dDiff < 0 ) ? LESS : MORE;
					}
					else
					{
						final double dMinX1 = pi1.getMinX( );
						final double dMinX2 = pi2.getMinX( );
						dDiff = dMinX1 - dMinX2;
						if ( !ChartUtil.mathEqual( dDiff, 0 ) )
						{
							return ( dDiff < 0 ) ? LESS : MORE;
						}
						else
						{
							final double dMaxX1 = pi1.getMaxX( );
							final double dMaxX2 = pi2.getMaxX( );
							dDiff = dMaxX1 - dMaxX2;
							if ( !ChartUtil.mathEqual( dDiff, 0 ) )
							{
								return ( dDiff < 0 ) ? LESS : MORE;
							}
							else
							{
								return EQUAL;
							}
						}
					}
				}

			}
			return EQUAL;
		}

		private final double getMinY( )
		{
			return _bo.getTop( );
		}

		private final double getMinX( )
		{
			return _bo.getLeft( );
		}

		private final double getMaxX( )
		{
			return _bo.getLeft( ) + _bo.getWidth( );
		}

		private final double getMaxY( )
		{
			return _bo.getTop( ) + _bo.getHeight( );
		}

		private final Location[] toLocationArray( )
		{
			final int n = _daXPoints.length;
			Location[] loa = new Location[n];
			for ( int i = 0; i < n; i++ )
			{
				loa[i] = LocationImpl.create( _daXPoints[i], _daYPoints[i] );
			}
			return loa;
		}
	}

	/**
	 * IDrawable
	 */
	private interface IDrawable
	{

		void draw( ) throws ChartException;

		Bounds getBounds( );
	}

	public class PieSlice
	{

		private boolean isExploded = true;
		private double startAngle;
		private double sliceLength;
		private double slicePecentage;
		private int categoryIndex;
		private DataPointHints dataPointHints;
		private double primitiveValue;

		private int quadrant = -1;

		private Location loPie;

		private Location loStart;

		private Location loEnd;

		private BoundingBox labelBounding = null;

		private Bounds bounds = null;
		private double w, h, xc, yc;

		PieSlice( double primitiveValue, DataPointHints dataPointHints,
				int categroyIndex )
		{
			this.primitiveValue = primitiveValue;
			this.dataPointHints = dataPointHints;
			this.categoryIndex = categroyIndex;
		}

		private PieSlice( )
		{
			// Used to clone
		}

		public Object clone( )
		{
			PieSlice slice = new PieSlice( );
			copyTo( slice );
			return slice;
		}

		public void copyTo( PieSlice slice )
		{
			slice.primitiveValue = primitiveValue;
			slice.dataPointHints = dataPointHints;
			slice.categoryIndex = this.categoryIndex;
			slice.setLabelLocation( loPie, loStart, loEnd );
			slice.setExploded( isExploded );
			slice.setStartAngle( startAngle );
			slice.setSliceLength( sliceLength );
			slice.setPercentage( slicePecentage );
			slice.setBounds( bounds );
			slice.labelBounding = labelBounding;
			slice.quadrant = quadrant;
		}

		public double getPrimitiveValue( )
		{
			return primitiveValue;
		}

		public DataPointHints getDataPointHints( )
		{
			return dataPointHints;
		}

		public String getDisplayValue( )
		{
			return dataPointHints.getDisplayValue( );
		}

		public double getStartAngle( )
		{
			return startAngle;
		}

		public double getSliceLength( )
		{
			return sliceLength;
		}

		public double getdMidAngle( )
		{
			return startAngle + sliceLength / 2;
		}

		public double getSlicePercentage( )
		{
			return slicePecentage;
		}

		public int getCategoryIndex( )
		{
			return categoryIndex;
		}

		public int getQuadrant( )
		{
			return quadrant;
		}

		public void setLabelLocation( double dX0, double dY0, double dX1,
				double dY1, double dX2, double dY2 )
		{
			setLabelLocation( LocationImpl.create( dX0, dY0 ),
					LocationImpl.create( dX1, dY1 ),
					LocationImpl.create( dX2, dY2 ) );
		}

		public void setLabelLocation( Location loPie, Location loStart,
				Location loEnd )
		{
			this.loPie = loPie;
			this.loStart = loStart;
			this.loEnd = loEnd;
		}

		public void setExploded( boolean isExploded )
		{
			this.isExploded = isExploded;
		}

		public void setStartAngle( double startAngle )
		{
			this.startAngle = startAngle;
		}

		public void setSliceLength( double newLength )
		{
			sliceLength = newLength;
		}

		public void setPercentage( double newPercentage )
		{
			slicePecentage = newPercentage;
		}

		public final BoundingBox getLabelBounding( )
		{
			return labelBounding;
		}

		public void removeLabelBounding( )
		{
			// Added to resolve empty category stacking (Bugzilla 197128)
			labelBounding = null;
		}

		/**
		 * @param loC
		 *            center coordinates of pie slice.
		 * @param loOffset
		 *            pie slice officeset for multiple dimension.
		 * @param sz
		 *            the width and height of pie slice.
		 * @param fi
		 *            the fill properties.
		 * @param iPieceType
		 * @throws ChartException
		 */
		private final void render( Location loC, Location loOffset, Size sz,
				Fill fi, int iPieceType ) throws ChartException
		{
			loC.translate( loOffset.getX( ) / 2d, loOffset.getY( ) / 2d );

			if ( isExploded && dExplosion != 0 )
			{
				// double dRatio = (double) d.width / d.height;
				double dMidAngleInRadians = Math.toRadians( getStartAngle( )
						+ getSliceLength( )
						/ 2d );
				double dSineThetaMid = ( Math.sin( dMidAngleInRadians ) );
				double dCosThetaMid = ( Math.cos( dMidAngleInRadians ) );
				double xDelta = ( dExplosion * dCosThetaMid );
				double yDelta = ( dExplosion * dSineThetaMid );

				loC.translate( xDelta, -yDelta );
			}

			Location loCTop = LocationImpl.create( loC.getX( )
					- loOffset.getX( ), loC.getY( ) - loOffset.getY( ) );
			double dAngleInRadians = Math.toRadians( getStartAngle( ) );
			double dSineThetaStart = Math.sin( dAngleInRadians );
			double dCosThetaStart = Math.cos( dAngleInRadians );
			dAngleInRadians = Math.toRadians( getStartAngle( )
					+ getSliceLength( ) );
			double dSineThetaEnd = Math.sin( dAngleInRadians );
			double dCosThetaEnd = Math.cos( dAngleInRadians );

			double xE = ( sz.getWidth( ) * dCosThetaEnd );
			double yE = ( sz.getHeight( ) * dSineThetaEnd );
			double xS = ( sz.getWidth( ) * dCosThetaStart );
			double yS = ( sz.getHeight( ) * dSineThetaStart );

			ArcRenderEvent are = null;
			if ( iPieceType == LOWER )
			{
				are = new ArcRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dataPointHints ) );
			}
			else
			{
				are = (ArcRenderEvent) ( (EventObjectCache) idr ).getEventObject( WrappedStructureSource.createSeriesDataPoint( ps,
						getDataPointHints( ) ),
						ArcRenderEvent.class );
			}
			are.setBackground( fi );
			liaEdges.setColor( getSliceOutline( fi ) );
			are.setOutline( liaEdges );
			are.setTopLeft( LocationImpl.create( loCTop.getX( ) - sz.getWidth( ),
					loCTop.getY( )
							- sz.getHeight( )
							+ ( iPieceType == LOWER ? dThickness : 0 ) ) );
			are.setWidth( sz.getWidth( ) * 2 );
			are.setHeight( sz.getHeight( ) * 2 );

			are.setStartAngle( startAngle );
			are.setAngleExtent( sliceLength );
			are.setStyle( ArcRenderEvent.SECTOR );
			idr.fillArc( are );

			if ( iPieceType == LOWER )
			{
				// DRAWN INTO A BUFFER FOR DEFERRED RENDERING
				double[] daXPoints = {
						loC.getX( ),
						loCTop.getX( ),
						loCTop.getX( ) + xE,
						loC.getX( ) + xE
				};
				double[] daYPoints = {
						loC.getY( ),
						loCTop.getY( ),
						loCTop.getY( ) - yE,
						loC.getY( ) - yE
				};
				deferFlatPlane( daXPoints,
						daYPoints,
						fi,
						loC.getX( ),
						loC.getX( ) + xE,
						dataPointHints );

				daXPoints = new double[]{
						loC.getX( ),
						loC.getX( ) + xS,
						loCTop.getX( ) + xS,
						loCTop.getX( )
				};
				daYPoints = new double[]{
						loC.getY( ),
						loC.getY( ) - yS,
						loCTop.getY( ) - yS,
						loCTop.getY( )
				};
				deferFlatPlane( daXPoints,
						daYPoints,
						fi,
						loC.getX( ),
						loC.getX( ) + xS,
						dataPointHints );

				daXPoints = new double[]{
						loC.getX( ) + xS,
						loCTop.getX( ) + xS,
						loCTop.getX( ) + xE,
						loC.getX( ) + xE
				};
				daYPoints = new double[]{
						loC.getY( ) - yS,
						loCTop.getY( ) - yS,
						loCTop.getY( ) - yE,
						loC.getY( ) - yE
				};

				final LineRenderEvent lreStartB2T = new LineRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dataPointHints ) );
				lreStartB2T.setStart( LocationImpl.create( loC.getX( ) + xS,
						loC.getY( ) - yS ) );
				lreStartB2T.setEnd( LocationImpl.create( loCTop.getX( ) + xS,
						loCTop.getY( ) - yS ) );
				final LineRenderEvent lreEndT2B = new LineRenderEvent( WrappedStructureSource.createSeriesDataPoint( ps,
						dataPointHints ) );
				lreEndT2B.setStart( LocationImpl.create( loCTop.getX( ) + xE,
						loCTop.getY( ) - yE ) );
				lreEndT2B.setEnd( LocationImpl.create( loC.getX( ) + xE,
						loC.getY( ) - yE ) );
				Bounds r2ddTop = BoundsImpl.create( loCTop.getX( )
						- sz.getWidth( ),
						loCTop.getY( ) - sz.getHeight( ),
						sz.getWidth( ) * 2,
						sz.getHeight( ) * 2 );
				Bounds r2ddBottom = BoundsImpl.create( loC.getX( )
						- sz.getWidth( ),
						loC.getY( ) - sz.getHeight( ),
						sz.getWidth( ) * 2,
						sz.getHeight( ) * 2 );

				registerCurvedSurface( r2ddTop,
						r2ddBottom,
						getStartAngle( ),
						getSliceLength( ),
						lreStartB2T,
						lreEndT2B,
						fi,
						dataPointHints,
						loC,
						loCTop,
						sz );
			}

			else if ( iPieceType == UPPER ) // DRAWN IMMEDIATELY
			{
				if ( ps.getSliceOutline( ) != null )
				{
					idr.drawArc( are );
				}

				if ( pie.isInteractivityEnabled( ) )
				{
					final EList elTriggers = ps.getTriggers( );
					if ( !elTriggers.isEmpty( ) )
					{
						final StructureSource iSource = WrappedStructureSource.createSeriesDataPoint( ps,
								dataPointHints );
						final InteractionEvent iev = (InteractionEvent) ( (EventObjectCache) idr ).getEventObject( iSource,
								InteractionEvent.class );
						Trigger tg;
						for ( int t = 0; t < elTriggers.size( ); t++ )
						{
							tg = TriggerImpl.copyInstance( (Trigger) elTriggers.get( t ) );
							pie.processTrigger( tg, iSource );
							iev.addTrigger( tg );
						}
						iev.setHotSpot( are );
						idr.enableInteraction( iev );
					}
				}
			}
		}

		private final void renderLabel( IDeviceRenderer idr, int iTextRenderType )
				throws ChartException
		{
			if ( labelBounding == null )
			{// Do not render if no bounding
				return;
			}
			if ( quadrant != -1 )
			{
				if ( iTextRenderType == TextRenderEvent.RENDER_TEXT_AT_LOCATION )
				{
					LineRenderEvent lre = (LineRenderEvent) ( (EventObjectCache) idr ).getEventObject( WrappedStructureSource.createSeriesDataPoint( ps,
							dataPointHints ),
							LineRenderEvent.class );
					lre.setLineAttributes( liaLL );
					lre.setStart( loPie );
					lre.setEnd( loStart );
					idr.drawLine( lre );

					lre = (LineRenderEvent) ( (EventObjectCache) idr ).getEventObject( WrappedStructureSource.createSeriesDataPoint( ps,
							dataPointHints ),
							LineRenderEvent.class );
					lre.setLineAttributes( liaLL );
					lre.setStart( loStart );
					lre.setEnd( loEnd );
					idr.drawLine( lre );
				}

				pie.renderLabel( WrappedStructureSource.createSeriesDataPoint( ps,
						dataPointHints ),
						TextRenderEvent.RENDER_TEXT_IN_BLOCK,
						laDataPoint,
						( quadrant == 1 || quadrant == 4 ) ? Position.RIGHT_LITERAL
								: Position.LEFT_LITERAL,
						loEnd,
						BoundsImpl.create( labelBounding.getLeft( ),
								labelBounding.getTop( ),
								labelBounding.getWidth( ),
								labelBounding.getHeight( ) ) );
			}
			else
			{
				pie.renderLabel( StructureSource.createSeries( ps ),
						TextRenderEvent.RENDER_TEXT_IN_BLOCK,
						laDataPoint,
						null,
						null,
						BoundsImpl.create( labelBounding.getLeft( ),
								labelBounding.getTop( ),
								labelBounding.getWidth( ),
								labelBounding.getHeight( ) ) );
			}
		}

		public boolean isLabelClipped( Bounds bo )
		{
			if ( labelBounding != null )
			{
				if ( labelBounding.getTop( ) < bo.getTop( ) )
				{
					return true;
				}
				if ( labelBounding.getLeft( ) < bo.getLeft( ) )
				{
					return true;
				}
				if ( labelBounding.getTop( ) + labelBounding.getHeight( ) > bo.getTop( )
						+ bo.getHeight( ) )
				{
					return true;
				}
				if ( labelBounding.getLeft( ) + labelBounding.getWidth( ) > bo.getLeft( )
						+ bo.getWidth( ) )
				{
					return true;
				}
			}
			return false;
		}

		public boolean isLabelOverlap( PieSlice sliceToCompare )
		{
			if ( sliceToCompare == null
					|| sliceToCompare == this
					|| sliceToCompare.labelBounding == null )
			{
				return false;
			}
			BoundingBox bb1 = labelBounding, bb2 = sliceToCompare.labelBounding;
			BoundingBox dHigh, dLow;
			// check which one is higher
			if ( bb1.getTop( ) < bb2.getTop( ) )
			{
				dHigh = bb1;
				dLow = bb2;
			}
			else
			{
				dHigh = bb2;
				dLow = bb1;
			}
			double dXHigh, dXLow, dYHigh, dYLow;
			if ( dHigh.getLeft( ) < dLow.getLeft( ) )
			{
				dXHigh = dHigh.getLeft( ) + dHigh.getWidth( );
				dYHigh = dHigh.getTop( ) + dHigh.getHeight( );
				dXLow = dLow.getLeft( );
				dYLow = dLow.getTop( );
				if ( dXHigh > dXLow && dYHigh > dYLow )
				{
					return true;
				}
			}
			else
			{
				dXHigh = dHigh.getLeft( );
				dYHigh = dHigh.getTop( ) + dHigh.getHeight( );
				dXLow = dLow.getLeft( ) + dLow.getWidth( );
				dYLow = dLow.getTop( );
				if ( dXHigh < dXLow && dYHigh > dYLow )
				{
					return true;
				}
			}
			return false;
		}

		public void setBounds( Bounds bo )
		{
			bounds = bo;
			w = bounds.getWidth( ) / 2 - dExplosion;
			h = bounds.getHeight( ) / 2 - dExplosion - dThickness / 2;
			xc = bounds.getLeft( ) + w + dExplosion;
			yc = bounds.getTop( ) + h + dExplosion + dThickness / 2;

			if ( ratio > 0 && w > 0 )
			{
				if ( h / w > ratio )
				{
					h = w * ratio;
				}
				else if ( h / w < ratio )
				{
					w = h / ratio;
				}
			}

			// detect invalid size.
			if ( w <= 0 || h <= 0 )
			{
				w = h = 1;
			}
		}

		private void computeLabelBoundOutside( LeaderLineStyle lls,
				double dLeaderLength )
		{
			int iLL = 0;
			double dLeaderTick = Math.max( dLeaderLength / 4,
					LEADER_TICK_MIN_SIZE * pie.getDeviceScale( ) );

			double dLeaderW = 0, dLeaderH = 0, dBottomLeaderW = 0, dBottomLeaderH = 0, dTopLeaderW = 0, dTopLeaderH = 0;
			Location center = LocationImpl.create( xc, yc - dThickness / 2 );
			Location depthCenter = LocationImpl.create( xc, yc ); // center in
			// the
			// middle of the depth
			double dX = 0;
			double dLeftSide = xc - dExplosion - w;
			double dRightSide = xc + dExplosion + w;

			if ( w > h )
			{
				dTopLeaderW = dLeaderTick;
				dTopLeaderH = dLeaderTick * ratio;
			}
			else
			{
				dTopLeaderH = dLeaderTick;
				dTopLeaderW = dLeaderTick / ratio;
			}
			double dMidAngleInDegrees = getdMidAngle( );
			double dMidAngleInRadians = Math.toRadians( -dMidAngleInDegrees );
			double dSineThetaMid = Math.sin( dMidAngleInRadians );
			double dCosThetaMid = Math.cos( dMidAngleInRadians );

			if ( dThickness > 0
					&& dMidAngleInDegrees > 180
					&& dMidAngleInDegrees < 360 )
			{
				double dTmpLeaderTick = Math.max( dThickness
						* dSineThetaMid
						+ 8
						* pie.getDeviceScale( ), dLeaderTick );
				if ( w > h )
				{
					dBottomLeaderW = dTmpLeaderTick;
					dBottomLeaderH = dTmpLeaderTick * ratio;
				}
				else
				{
					dBottomLeaderH = dTmpLeaderTick;
					dBottomLeaderW = dTmpLeaderTick / ratio;
				}

				dLeaderW = dBottomLeaderW;
				dLeaderH = dBottomLeaderH;
			}
			else
			{
				dLeaderW = dTopLeaderW;
				dLeaderH = dTopLeaderH;
			}

			double xDelta1, yDelta1, xDelta2, yDelta2;
			if ( isExploded )
			{
				xDelta1 = ( w + dExplosion ) * dCosThetaMid;
				yDelta1 = ( h + dExplosion ) * dSineThetaMid;
				xDelta2 = ( w + dLeaderW + dExplosion ) * dCosThetaMid;
				yDelta2 = ( h + dLeaderH + dExplosion ) * dSineThetaMid;
			}
			else
			{
				xDelta1 = ( w ) * dCosThetaMid;
				yDelta1 = ( h ) * dSineThetaMid;
				xDelta2 = ( w + dLeaderW ) * dCosThetaMid;
				yDelta2 = ( h + dLeaderH ) * dSineThetaMid;
			}

			if ( lls == LeaderLineStyle.STRETCH_TO_SIDE_LITERAL )
			{
				if ( dMidAngleInDegrees >= 90 && dMidAngleInDegrees < 270 )
				{
					dX = dLeftSide - dLeaderW * 1.5;
					iLL = IConstants.LEFT;
				}
				else
				{
					dX = dRightSide + dLeaderW * 1.5;
					iLL = IConstants.RIGHT;
				}
			}
			else if ( lls == LeaderLineStyle.FIXED_LENGTH_LITERAL )
			{
				if ( dMidAngleInDegrees > 90 && dMidAngleInDegrees < 270 )
				{
					dX = center.getX( ) + xDelta2 - dLeaderLength;
					if ( dLeaderLength > 0 )
					{
						iLL = IConstants.LEFT;
					}
					else
					{
						if ( dMidAngleInDegrees < 135 )
						{
							iLL = IConstants.TOP;
						}
						else if ( dMidAngleInDegrees < 225 )
						{
							iLL = IConstants.LEFT;
						}
						else if ( dMidAngleInDegrees < 270 )
						{
							iLL = IConstants.BOTTOM;
						}
						else
							assert false;
					}
				}
				else
				{
					dX = center.getX( ) + xDelta2 + dLeaderLength;
					if ( dLeaderLength > 0 )
					{
						iLL = IConstants.RIGHT;
					}
					else
					{
						if ( dMidAngleInDegrees <= 45 )
						{
							iLL = IConstants.RIGHT;
						}
						else if ( dMidAngleInDegrees > 45
								&& dMidAngleInDegrees <= 90 )
						{
							iLL = IConstants.TOP;
						}
						else if ( dMidAngleInDegrees <= 315
								&& dMidAngleInDegrees >= 270 )
						{
							iLL = IConstants.BOTTOM;
						}
						else if ( dMidAngleInDegrees > 315 )
						{
							iLL = IConstants.RIGHT;
						}
						else
							assert false;
					}

				}
			}
			else
			{
				// SHOULD'VE ALREADY THROWN THIS EXCEPTION PREVIOUSLY
			}

			Location relativeCenter;
			if ( dMidAngleInDegrees > 0 && dMidAngleInDegrees < 180 )
			{
				relativeCenter = center;
			}
			else
			{
				relativeCenter = depthCenter;
			}
			setLabelLocation( relativeCenter.getX( ) + xDelta1,
					relativeCenter.getY( ) + yDelta1,
					relativeCenter.getX( ) + xDelta2,
					relativeCenter.getY( ) + yDelta2,
					dX,
					relativeCenter.getY( ) + yDelta2 );
			laDataPoint.getCaption( ).setValue( getDisplayValue( ) );
			labelBounding = Methods.computeBox( xs,
					iLL,
					laDataPoint,
					dX,
					relativeCenter.getY( ) + yDelta2 );
			// NEEDED FOR COMPUTING DYNAMIC REPOSITIONING LIMITS
			if ( dMidAngleInDegrees >= 0 && dMidAngleInDegrees < 90 )
			{
				quadrant = 1;
			}
			if ( dMidAngleInDegrees >= 90 && dMidAngleInDegrees < 180 )
			{
				quadrant = 2;
			}
			if ( dMidAngleInDegrees >= 180 && dMidAngleInDegrees < 270 )
			{
				quadrant = 3;
			}
			else
			{
				quadrant = 4;
			}
		}

		private void computeLabelBoundInside( )
		{
			double dMidAngleInRadians = Math.toRadians( -getdMidAngle( ) );
			double dSineThetaMid = Math.sin( dMidAngleInRadians );
			double dCosThetaMid = Math.cos( dMidAngleInRadians );
			double xDelta, yDelta;
			if ( isExploded )
			{
				xDelta = ( ( w / 1.5d + dExplosion ) * dCosThetaMid );
				yDelta = ( ( h / 1.5d + dExplosion ) * dSineThetaMid );
			}
			else
			{
				xDelta = ( ( w / 1.5d ) * dCosThetaMid );
				yDelta = ( ( h / 1.5d ) * dSineThetaMid );
			}
			laDataPoint.getCaption( ).setValue( getDisplayValue( ) );
			labelBounding = Methods.computeBox( xs,
					IConstants.LEFT/* DONT-CARE */,
					laDataPoint,
					0,
					0 );
			labelBounding.setLeft( xc + xDelta - labelBounding.getWidth( ) / 2 );
			labelBounding.setTop( yc
					- dThickness
					/ 2
					+ yDelta
					- labelBounding.getHeight( )
					/ 2 );
		}
	}
}